import { buildASTSchema, buildSchema, isSchema, } from 'graphql';
import { createYoga, } from 'graphql-yoga';
import { useSupergraph } from '@graphql-mesh/fusion-runtime';
// eslint-disable-next-line import/no-extraneous-dependencies
import { DefaultLogger, getHeadersObj } from '@graphql-mesh/utils';
import { getStitchedSchemaFromSupergraphSdl } from '@graphql-tools/federation';
import { getDocumentNodeFromSchema } from '@graphql-tools/utils';
import { isPromise } from '@whatwg-node/server';
export function createMeshHTTPHandler(config) {
    let handlerImportFn;
    if (config.handlers != null) {
        if (typeof config.handlers === 'function') {
            handlerImportFn = config.handlers;
        }
        else {
            handlerImportFn = (handlerName) => config.handlers[handlerName];
        }
    }
    else {
        handlerImportFn = defaultHandlerImport;
    }
    let fetchAPI = config.fetchAPI;
    // eslint-disable-next-line prefer-const
    let logger;
    function defaultHandlerImport(handlerName) {
        const omnigraphPackageName = `@omnigraph/${handlerName}`;
        logger.info(`Loading ${omnigraphPackageName}`);
        return import(omnigraphPackageName);
    }
    const supergraphSpec = config.spec || 'fusion';
    let supergraphYogaPlugin;
    if (supergraphSpec === 'fusion') {
        supergraphYogaPlugin = useSupergraph({
            getSupergraph() {
                if (typeof config.supergraph === 'function') {
                    return config.supergraph();
                }
                if (isSchema(config.supergraph)) {
                    return config.supergraph;
                }
                if (typeof config.supergraph === 'string') {
                    let url = config.supergraph;
                    if (url.startsWith('.')) {
                        url = `file:///${url}`;
                    }
                    yoga.logger.info(`Fetching Supergraph from ${url}`);
                    return fetchAPI
                        .fetch(config.supergraph, {
                        headers: config.schemaHeaders,
                    })
                        .then(res => res.text())
                        .then(schemaString => buildSchema(schemaString, {
                        assumeValid: true,
                        assumeValidSDL: true,
                        noLocation: true,
                    }));
                }
                return buildASTSchema(config.supergraph, {
                    assumeValid: true,
                    assumeValidSDL: true,
                });
            },
            getExecutorFromHandler(handlerName, handlerOptions, getSubgraph) {
                function handleImportResult(importRes) {
                    const getSubgraphExecutor = importRes.getSubgraphExecutor;
                    if (!getSubgraphExecutor) {
                        logger.error(`getSubgraphExecutor is not exported from the handler: ${handlerName}`);
                        throw new Error(`getSubgraphExecutor is not exported from the handler: ${handlerName}`);
                    }
                    return getSubgraphExecutor({
                        getSubgraph,
                        options: handlerOptions,
                    });
                }
                const importRes$ = handlerImportFn(handlerName);
                if (isPromise(importRes$)) {
                    return importRes$.then(handleImportResult);
                }
                return handleImportResult(importRes$);
            },
            polling: config.polling,
        });
    }
    else if (supergraphSpec === 'federation') {
        let supergraph;
        // eslint-disable-next-line no-inner-declarations
        function getAndSetSupergraph() {
            function handleSupergraphInput(supergraphSdl) {
                supergraph = getStitchedSchemaFromSupergraphSdl({
                    supergraphSdl: isSchema(supergraphSdl)
                        ? getDocumentNodeFromSchema(supergraphSdl)
                        : supergraphSdl,
                });
            }
            if (typeof config.supergraph === 'function') {
                const supergraph$ = config.supergraph();
                if (isPromise(supergraph$)) {
                    return supergraph$.then(handleSupergraphInput);
                }
                return handleSupergraphInput(supergraph$);
            }
            if (isPromise(config.supergraph)) {
                return config.supergraph.then(handleSupergraphInput);
            }
            return handleSupergraphInput(config.supergraph);
        }
        let initialSupergraph$;
        supergraphYogaPlugin = {
            onRequestParse() {
                return {
                    onRequestParseDone() {
                        initialSupergraph$ ||= getAndSetSupergraph();
                        return initialSupergraph$;
                    },
                };
            },
            onEnveloped({ setSchema }) {
                setSchema(supergraph);
            },
            invalidateSupergraph() {
                return getAndSetSupergraph();
            },
        };
    }
    const yoga = createYoga({
        fetchAPI: config.fetchAPI,
        logging: config.logging == null ? new DefaultLogger() : config.logging,
        plugins: [...(config.plugins || []), supergraphYogaPlugin],
        context({ request, req, connectionParams }) {
            // Maybe Node-like environment
            if (req?.headers) {
                return {
                    fetch: fetchAPI.fetch,
                    logger,
                    headers: getHeadersObj(req.headers),
                    connectionParams,
                };
            }
            // Fetch environment
            if (request?.headers) {
                return {
                    fetch: fetchAPI.fetch,
                    logger,
                    headers: getHeadersObj(request.headers),
                    connectionParams,
                };
            }
            return {};
        },
        cors: config.cors,
        graphiql: config.graphiql,
        batching: config.batching,
    });
    fetchAPI ||= yoga.fetchAPI;
    logger = yoga.logger;
    Object.defineProperty(yoga, 'invalidateSupergraph', {
        value: supergraphYogaPlugin.invalidateSupergraph,
        configurable: true,
    });
    return yoga;
}
